# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from . import (
    E_APP_ARG_BADOPT, E_APP_ARG_DUPEOPT, E_APP_MISSING,
    E_APP_ARG_FEW, E_APP_ARG_MANY, E_APP_ARG_PIPEDELIM,
    E_APP_ARG_IFCONST, E_APP_ARG_IFEMPTY, E_APP_ARG_IFSTYLE,
    W_APP_BALANCE, W_APP_BAD_CASE)
from ..variable import Var, strjoin, variable_check_balance


# TODO: instead of where, we should pass a context object that context
# could include a where, and also a list of classes where we can store
# significant info like "label" when there is a goto/gosub that jumps to
# one.


class AppArg(object):
    def __init__(self, name):
        self.name = name
        # The following are set afterwards.
        self.argno = None
        self.app = None

    def validate(self, arg, where):
        pass


class AppOptions(AppArg):
    def __init__(self, options):
        super().__init__('options')
        self.options = options

    def validate(self, arg, where):
        str_options = [i for i in arg if isinstance(i, str)]
        bad_options = [i for i in str_options if i not in self.options]
        if bad_options:
            E_APP_ARG_BADOPT(where, argno=self.argno, app=self.app,
                             opts=''.join(bad_options))
        if len(str_options) != len(set(str_options)):
            E_APP_ARG_DUPEOPT(where, argno=self.argno, app=self.app,
                              opts=arg)


class AppBase(object):
    @property
    def name(self):
        return self.__class__.__name__

    @property
    def module(self):
        return self.__module__.rsplit('.', 1)[-1]

    def __call__(self, data, where, jump_destinations):
        """
        Subclasses may implement something that populates
        jump_destinations.
        """
        try:
            self.check_balance(data)
        except ValueError:
            W_APP_BALANCE(where, data=str(data))

    def check_availability(self, supplied_name, where):
        if self.name == supplied_name:
            pass
        elif self.name == 'Unknown':
            E_APP_MISSING(where, app=supplied_name)
        else:
            W_APP_BAD_CASE(where, app=supplied_name, proper=self.name)

    @staticmethod
    def check_balance(data):
        variable_check_balance(data)

    @staticmethod
    def separate_args(data, delimiter=',', remove_quotes_backslashes=True):
        # #define AST_STANDARD_APP_ARGS(...) => separate_args(',', True)
        # #define AST_STANDARD_RAW_ARGS(...) => separate_args(',', False)
        # SOURCE: main/app.c -- __ast_app_separate_args()
        #
        # TODO: we should separate args using a more sensible approach as
        # well, so we can warn on inconsistencies.
        brackets = 0
        parens = 0
        quotes = False
        skipnext = False

        ret = [[]]
        start = 0
        for i, char in enumerate(data):
            if isinstance(char, Var):
                # It's possible that we're looping over a Var variable.
                # In that case we'll have to assume the Var itself
                # contains no separators that we might be interested in.
                # E.g. we expect you do *not* do this:
                # ``Set(var=mailbox@context,s)``
                # ``VoiceMail(${var})``
                # But we expect you to do this:
                # ``Set(mailbox=mailbox@context)``
                # ``VoiceMail(${mailbox},s)``
                # Skip it?
                # #ret[-1].extend(data[start:i])  # skip char
                # #start = i + 1
                # Don't skip it, we want the returned value.
                pass
            elif skipnext:
                skipnext = False
            elif char == '[':
                brackets += 1
            elif char == ']':
                if brackets:
                    brackets -= 1
            elif char == '(':
                parens += 1
            elif char == ')':
                if parens:
                    parens -= 1
            elif char == '"' and delimiter != '"':
                quotes = not quotes
                if remove_quotes_backslashes:
                    ret[-1].extend(data[start:i])
                    start = i + 1
            elif char == '\\':
                if remove_quotes_backslashes:
                    ret[-1].extend(data[start:i])
                    start = i + 1
                skipnext = True
            elif char == delimiter and not (brackets or parens or quotes):
                ret[-1].extend(data[start:i])
                start = i + 1
                ret.append([])  # start on next arg

        # Append leftover args.
        ret[-1].extend(data[start:])

        # Squash args.
        squashed = []
        for letters in ret:
            letters = list(strjoin(letters))  # join sequences of strings
            if len(letters) == 1:
                squashed.append(letters[0])
            else:
                squashed.append(Var.join(letters))

        return squashed


class DelimitedArgsMixin(object):
    def __init__(self, arg_delimiter=',', arg_raw=False, **kwargs):
        self._arg_delimiter = arg_delimiter
        self._arg_raw = arg_raw

    def split_args(self, data, where):
        return self.separate_args(
            data, delimiter=self._arg_delimiter,
            remove_quotes_backslashes=self._arg_raw)

    def __call__(self, data, where, jump_destinations):
        return self.split_args(data, where)


class MinMaxArgsMixin(object):
    def __init__(self, min_args=None, max_args=None, **kwargs):
        super().__init__(**kwargs)
        assert (min_args is None or max_args is None or
                min_args <= max_args)
        self._min_args = min_args
        self._max_args = max_args

    def split_args(self, data, where):
        args = super().split_args(data, where)

        if self._min_args is not None and (
                (len(args) < self._min_args) or
                (self._min_args == 1 and len(args) == 1 and not args[0])):
            E_APP_ARG_FEW(where, app=self.name, min_args=self._min_args)
        elif self._max_args is not None and (
                (len(args) > self._max_args)):
            E_APP_ARG_MANY(where, app=self.name, max_args=self._max_args)

        return args


class AppArgsMixin(MinMaxArgsMixin):
    def __init__(self, args=None, **kwargs):
        assert args, 'AppArgsMixin wants some args for {!r}'.format(
            self.__class__.__name__)

        super().__init__(max_args=len(args), **kwargs)

        # Set the arguments and pimp them a bit with more info.
        self._args = args
        for i, arg in enumerate(self._args):
            arg.argno = i + 1
            arg.app = self.name

    def split_args(self, data, where):
        args = super().split_args(data, where)

        for i, arg in enumerate(args):
            if len(self._args) > i:
                self._args[i].validate(arg, where)

        return args


class NoPipeDelimiterMixin(object):
    # TODO: we should not use this for the System() app,
    # nor for the SHELL() function,
    # and it would be improved if we checked for surrounding quotes
    # which clearly show that the user isn't doing anything unintended.

    def split_args(self, data, where):
        args = super().split_args(data, where)
        if len(args) == 1 and '|' in args[0]:
            E_APP_ARG_PIPEDELIM(where, app=self.name, data=data)

        return args


class App(NoPipeDelimiterMixin, AppArgsMixin, DelimitedArgsMixin, AppBase):
    """
    The App takes an optional args=[AppArg(...), ...] and max_args=INT,
    and checks the values if available.
    """
    pass


class IfStyleApp(DelimitedArgsMixin, AppBase):
    """
    The IfStyleApp takes the form <cond>?<iftrue>:<iffalse>.
    """
    def __init__(self, **kwargs):
        super().__init__(arg_delimiter='?', arg_raw=True, **kwargs)

    def split_args(self, data, where):
        args = super().split_args(data, where)

        if len(args) >= 2:
            if len(args) > 2:
                # This is no good either, with too many results.
                E_APP_ARG_MANY(where, app=self.name, data=data, max_args=2)

            # BEWARE: GosubIf and ExecIf use this method, while GotoIf
            # uses a simpler split. In the majority of cases the result
            # would be the same (barring crazy backslash usage).
            cond = args[0]
            actions = self.separate_args(
                args[1], delimiter=':', remove_quotes_backslashes=False)
            if len(actions) == 1:
                iftrue, iffalse = actions[0], None
            else:
                # We simply drop excess args here.
                assert len(actions) >= 2, actions
                iftrue, iffalse = actions[0], actions[1]

        else:
            assert len(args) == 1, args

            # Attempt to separate the arguments the old fashioned way.
            args = self.separate_args(
                data, ',', remove_quotes_backslashes=False)
            cond = args[0]
            iftrue = args[1:]
            if len(iftrue):
                E_APP_ARG_IFSTYLE(where, app=self.name, data=data,
                                  cond=args[0], args=args[1:])
            else:
                E_APP_ARG_FEW(where, app=self.name, data=data, min_args=2)

            # Observe that iftrue is already split this time.
            cond, iftrue, iffalse = args[0], args[1:], None

        # Trim space from the sides of cond.
        cond = cond.strip()
        if cond == '':
            E_APP_ARG_IFEMPTY(where, app=self.name, data=data)

        return cond, iftrue, iffalse


class VarCondIfStyleApp(IfStyleApp):
    """
    Like IfStyleApp it takes the form <cond>?<iftrue>:<iffalse>, but
    the condition cannot contain constants.

    Use this for GotoIf/ExecIf.
    """
    def split_args(self, data, where):
        cond, iftrue, iffalse = super().split_args(data, where)

        # Check whether cond is a constant.
        if (cond != '' and  # already checked with E_APP_ARG_IFEMPTY
                (isinstance(cond, str) or
                 any(isinstance(i, str) for i in cond))):
            # NOTE: The resultant value is parsed with sscanf, and the
            # remainder is discarded. For example, something that
            # evaluates to "1&0" is parsed as "1" => "$[1=1]&$[0=0]"
            # (which should have been written as "$[1=1&0=0]") would
            # evaluate to True.
            E_APP_ARG_IFCONST(where, app=self.name, data=data,
                              cond=cond)

        return cond, iftrue, iffalse



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re

from .config import (
    E_CONF_KEY_DUPE, E_CONF_KEY_INVALID, W_CONF_KEY_LATE,
    ConfigAggregator, Context)
from .defines import ErrorDef, WarningDef
from .func.base import FuncBase
from .varfun import FuncLoader

try:
    import sqlparse
except ImportError:
    sqlparse = None  # optional


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_ODBC_NOSQL(ErrorDef):
        message = ('a readsql and/or writesql (and insertsql) statement is '
                   'required')

    class W_ODBC_BADTOKENS(WarningDef):
        message = ('odbc function {name!r} should be uppercase '
                   'with A-Z0-9_ only')

    class W_ODBC_LEGACY(WarningDef):
        message = ('read and write are legacy, use readsql and writesql '
                   '(and insertsql) instead')

    class W_ODBC_TAILSEMI(WarningDef):
        message = 'tailing semicolon is unnecessary'


class FuncOdbcFunc(FuncBase):
    @property
    def module(self):
        # This would be correct automatically, since this module is
        # called func_odbc, but it doesn't hurt to be explicit.
        return 'func_odbc'


class OdbcQuery(object):
    @classmethod
    def from_query(cls, query):
        return cls(query.strip())

    def __init__(self, query):
        self._query = query

    def _require_sqlparse(self):
        if not sqlparse:
            raise ImportError('sqlparse not found; this code path requires it')

    def format_sql(self):
        self._require_sqlparse()
        return sqlparse.format(
            self._query, keyword_case='upper', identifier_case=None,
            reindent=True, indent_tabs=False, indent_width=4,
            wrap_after=72)

    def select_columns(self):
        self._require_sqlparse()
        parsed = sqlparse.parse(self._query)
        assert len(parsed) == 1, parsed  # expected 1 statement
        parsed = parsed[0]

        if not parsed.tokens[0].match(sqlparse.tokens.Keyword.DML, 'SELECT'):
            # TODO: warn that we're not doing a SELECT?
            return None

        identifiers = self._tokens_to_identifiers(parsed.tokens)

# See also this: https://github.com/andialbrecht/sqlparse/issues/157
        tokens = [i for i in parsed.tokens]
        tables = []
        for idx, token in enumerate(tokens):
            if (token.is_keyword and (
                    token.normalized in ('FROM', 'JOIN') or
                    token.normalized.endswith(' JOIN'))):
                assert tokens[idx + 1].is_whitespace, (idx, tokens)
                table_identifier = tokens[idx + 2]
                if table_identifier.is_group():
                    # table_name <WS> table_alias
                    assert len(table_identifier.tokens) in (1, 3), (
                        idx, tokens, table_identifier.tokens)
                    if len(table_identifier.tokens) == 3:
                        assert table_identifier.tokens[1].is_whitespace, (
                            idx, tokens, table_identifier.tokens)
                    table_identifier = table_identifier.tokens[0]

                assert table_identifier.ttype == sqlparse.tokens.Token.Name, (
                        idx, tokens, table_identifier.tokens)
                tables.append(table_identifier.value)
        print('TABLES', tables)

        return self._identifiers_to_columns(identifiers)

    def _tokens_to_identifiers(self, tokens):
        # SELECT <ws>
        assert len(tokens) >= 3, len(tokens)
        assert tokens[1].match(
            sqlparse.tokens.Whitespace, '\s+', regex=True)

        # SELECT <ws> <identifier[list]> [<ws> FROM]
        if len(tokens) == 3 or (
                len(tokens) >= 5 and
                tokens[4].match(sqlparse.tokens.Keyword, 'FROM')):
            return self._token_to_identifiers(tokens[2])

        # SELECT <ws> <keyword> <identifier[list]> [<ws> FROM] ?
        raise NotImplementedError(
            'Unparenthesized leading keyword in SQL SELECT columns.\n'
            '\n'
            'sqlparse 0.1.18-1 parses e.g.:\n'
            '  SELECT NOT column FROM table\n'
            'into:\n'
            '  <SELECT> <NOT> <IDENTIFIER[LIST]> <FROM> ..\n'
            'Please add parentheses like this:\n'
            '  SELECT (NOT column) AS ncol FROM table\n'
            'so it will parse like this:\n'
            '  <SELECT> <IDENTIFIER[LIST]> <FROM> ..\n'
            '\n'
            'Problematic query:\n'
            '  {}'.format(self._query))

    def _token_to_identifiers(self, token):
        # token is an identifier/function
        if isinstance(token, (
                sqlparse.sql.Function, sqlparse.sql.Identifier)):
            return [token]

        # token is an identifierlist
        if isinstance(token, sqlparse.sql.IdentifierList):
            return token.get_identifiers()

        raise TypeError('Unexpected {!r}: {!r} in query ({!r})'.format(
            type(token).__name__, token, self._query))

    def _identifiers_to_columns(self, identifiers):
        return [token.get_alias() or str(token)
                for token in identifiers]


class OdbcFunction(Context):
    ORDER = ('prefix', 'synopsis', 'dsn', 'escapecommas', 'mode', 'rowlimit',
             # 'insertsql' is used when the writesql update returned 0
             # updated rows.
             'readhandle', 'readsql', 'writehandle', 'writesql', 'insertsql',
             'read', 'write')  # legacy

    def get_name(self):
        return self.name

    def get_function_name(self):
        return '{}_{}'.format(self.get_prefix(), self.get_name())

    def get_prefix(self):
        return self.get_variable('prefix') or 'ODBC'

    def get_variable(self, variable):
        for var in self._varsets:
            if var.variable == variable:
                return var.value
        return None

    def get_queries(self):
        rquery = self.get_variable('readsql') or self.get_variable('read')
        wquery = self.get_variable('writesql') or self.get_variable('write')
        iquery = self.get_variable('insertsql')
        queries = {}
        if rquery:  # x=${FUNC()}
            queries['read'] = OdbcQuery.from_query(rquery)
        if wquery:  # FUNC()=x
            queries['write'] = OdbcQuery.from_query(wquery)
        if iquery:  # FUNC()=x if 'write' query updated 0 rows
            queries['insert'] = OdbcQuery.from_query(iquery)
        return queries

    def has_variable(self, variable):
        return self.get_variable(variable) is not None

    def add(self, varset):
        if self.check_order(self._varsets, varset):

            # Check for trailing semicolons.
            if varset.variable in (
                    'readsql', 'writesql', 'insertsql', 'read', 'write'):
                if varset.variable in ('read', 'write'):
                    W_ODBC_LEGACY(varset.where)
                if varset.value.endswith(';'):
                    W_ODBC_TAILSEMI(varset.where)
            # TODO:
            # - check for usage/missing ARGn/VALn (only VALn in
            #   write/insert)
            # - properly formatted ${SQL_ESC(${ARG4})}
            # - prefix should be uppercase
            # - check SQL syntax?

            return super().add(varset)

    def update(self, context):
        raise NotImplementedError('when? how? what?')

    def check_order(self, set_, new):
        try:
            new_pos = self.ORDER.index(new.variable)
        except ValueError:
            E_CONF_KEY_INVALID(new.where, key=new.variable)
            return False

        if set_:
            used_poses = [self.ORDER.index(i.variable) for i in set_]
            if new_pos in used_poses:
                # func_odbc.c uses ast_variable_retrieve (checked on jan
                # 2016 in Asterisk 13), so it uses the first key, not
                # the last.
                previous = [i for i in set_ if i.variable == new.variable][0]
                E_CONF_KEY_DUPE(new.where, key=new.variable, previous=previous)
                return False

            if new_pos < used_poses[-1]:
                # We expected this key sooner.
                W_CONF_KEY_LATE(new.where, key=new.variable)

        return True

    def finalize(self):
        # TODO:
        # - Check that we have not both read and readsql (and friends)
        # - Check that there is a dsn or a *handle for every *sql.
        # - Check existence of synopsis/syntax
        # - Note: syntax=(empty) does not work, use syntax=void or
        #   syntax=arg1,arg2 (syntax check should match comma's to
        #   ARGn count)

        if not self.finalize_check_any_read_write():
            return
        self.finalize_register()

    def finalize_check_any_read_write(self):
        if not (self.has_variable('read') or self.has_variable('write') or
                self.has_variable('readsql') or self.has_variable('writesql')):
            E_ODBC_NOSQL(self.where)
            return False
        return True

    def finalize_register(self):
        # Register the custom function.
        odbc_function_class = type(
            self.get_function_name(), (FuncOdbcFunc,), {})
        FuncLoader().register(odbc_function_class())


class FuncOdbcAggregator(ConfigAggregator):
    """
    TODO: This should instead be split up into a simple config parser
    which also handles inheritance and a really short func_odbc
    aggregator.
    """
    # Allow A_B_C and ABC, but not _A or B_ or AB__C.
    FUNC_NAME_RE = re.compile(r'^[A-Z0-9](_?[A-Z0-9])*$')

    def on_yield(self):
        for function in super().on_yield():
            function.finalize()
            yield function

        # TODO: - Check for dupe functions/contexts.  jan2016 check in
        # pbx.c says that duplicate functions do not get registered =>
        # first one wins

    def on_context(self, context):
        if not self.FUNC_NAME_RE.match(context.name):
            W_ODBC_BADTOKENS(context.where, name=context.name)

        context = OdbcFunction.from_context(context)
        return super().on_context(context)

    def on_varset(self, varset):
        assert not varset.arrow  # W_ARROW
        return super().on_varset(varset)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2018  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import E_APP_ARG_BADOPT, App, AppArg


class AppOrExten(AppArg):
    def validate(self, arg, where):
        if arg not in ('app', 'exten'):
            E_APP_ARG_BADOPT(
                where, argno=self.argno, app=self.app, opts=arg)


class Originate(App):
    def __init__(self):
        super().__init__(
            # arg1 means Application-name or Context
            args=[AppArg('tech_data'), AppOrExten('type'), AppArg('arg1'),
                  AppArg('arg2'), AppArg('arg3'), AppArg('timeout')],
            min_args=3)


def register(app_loader):
    app_loader.register(Originate())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import (
    AppBase, DelimitedArgsMixin, IfStyleApp, MinMaxArgsMixin,
    NoPipeDelimiterMixin, VarCondIfStyleApp)

from asterisklint.varfun import VarLoader


class _Builtin(object):
    @property
    def module(self):
        return '<builtin>'


class BuiltinAppBase(_Builtin, AppBase):
    pass


class Answer(BuiltinAppBase):
    pass


class Background(BuiltinAppBase):
    # NOTE: It is called "BackGround" in Asterisk, but that makes my
    # eyes sore. But, especially for those that prefer the "official"
    # syntax, we allow the first occurrence of the app to define the
    # future spelling.
    chosen_spelling = None

    @property
    def name(self):
        return self.chosen_spelling or super().name

    def check_availability(self, supplied_name, where):
        if not self.chosen_spelling and supplied_name in (
                'Background', 'BackGround'):
            self.chosen_spelling = supplied_name

        return super().check_availability(supplied_name, where)


class Busy(BuiltinAppBase):
    pass


class Congestion(BuiltinAppBase):
    pass


class Goto(_Builtin, NoPipeDelimiterMixin, MinMaxArgsMixin, DelimitedArgsMixin,
           AppBase):
    def __init__(self, **kwargs):
        super().__init__(min_args=1, max_args=3, **kwargs)

    def __call__(self, data, where, jump_destinations):
        args = super().__call__(data, where, jump_destinations)

        jumpdest = args[:]
        while len(jumpdest) != 3:
            jumpdest.insert(0, None)
        assert len(jumpdest) == 3
        jump_destinations.append(tuple(jumpdest))

        return args


class GotoIf(_Builtin, VarCondIfStyleApp):
    def __call__(self, data, where, jump_destinations):
        cond, iftrue, iffalse = super().__call__(
            data, where, jump_destinations)

        for args in (iftrue, iffalse):
            if args:
                jumpdest = self.separate_args(args)
                while len(jumpdest) != 3:
                    jumpdest.insert(0, None)
                assert len(jumpdest) == 3
                jump_destinations.append(tuple(jumpdest))

        return cond, iftrue, iffalse


class GotoIfTime(_Builtin, IfStyleApp):
    pass


class ExecIfTime(_Builtin, IfStyleApp):
    pass


class Hangup(BuiltinAppBase):
    pass


class NoOp(BuiltinAppBase):
    pass


class Proceeding(BuiltinAppBase):
    pass


class Progress(BuiltinAppBase):
    pass


class ResetCDR(BuiltinAppBase):
    pass


class Ringing(BuiltinAppBase):
    pass


class SayNumber(BuiltinAppBase):
    pass


class SayDigits(BuiltinAppBase):
    pass


class SayAlpha(BuiltinAppBase):
    pass


class SayPhonetic(BuiltinAppBase):
    pass


class SetAMAFlags(BuiltinAppBase):
    pass


class SetGlobalVar(BuiltinAppBase):
    pass


class Set(BuiltinAppBase):
    def __call__(self, data, where, jump_destinations):
        ret = VarLoader().parse_assignment(data, where)
        super().__call__(data, where, jump_destinations)
        return ret


class ImportVar(BuiltinAppBase):
    pass


class Wait(BuiltinAppBase):
    pass


class WaitExten(BuiltinAppBase):
    pass


def register(app_loader):
    for app in (
            Answer, Background, Busy,
            Congestion, Goto, GotoIf,
            GotoIfTime, ExecIfTime,
            Hangup, NoOp, Proceeding,
            Progress, ResetCDR, Ringing,
            SayNumber, SayDigits, SayAlpha,
            SayPhonetic, SetAMAFlags,
            SetGlobalVar, Set,
            ImportVar, Wait,
            WaitExten):
        app_loader.register(app())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Show which modules, apps and functions are used by the dialplan. Takes
'extensions.conf' as argument.
"""
from collections import defaultdict

from asterisklint import FileDialplanParser
from asterisklint.app import E_APP_MISSING
from asterisklint.application import AppLoader
from asterisklint.defines import MessageDefManager
from asterisklint.mainutil import (
    MainBase, UniqueStore, load_func_odbc_functions)
from asterisklint.varfun import E_FUNC_MISSING, FuncLoader


class Aggregator(object):
    """
    Collector of all E_APP_MISSING, E_FUNC_MISSING errors. We want to
    know about these things.
    """
    def __init__(self):
        E_APP_MISSING.add_callback(self.on_e_app_missing)
        E_FUNC_MISSING.add_callback(self.on_e_func_missing)
        self.unknown_apps = set()
        self.unknown_funcs = set()

    def on_e_app_missing(self, message):
        self.unknown_apps.add(message.fmtkwargs['app'])

    def on_e_func_missing(self, message):
        self.unknown_funcs.add(message.fmtkwargs['func'])


class Main(MainBase):
    def create_argparser(self, argparser_class):
        parser = argparser_class(
            description=(
                "Show which modules, apps and functions are used by "
                "the dialplan. "
                "Useful when you use autoload=no in your modules.conf. "
                "Beware that you do need more modules than just these "
                "listed."))
        parser.add_argument(
            'dialplan', metavar='EXTENSIONS_CONF', nargs='?',
            default='./extensions.conf',
            help='path to extensions.conf')
        parser.add_argument(
            '--func-odbc', metavar='FUNC_ODBC_CONF', action=UniqueStore,
            help="path to func_odbc.conf, will be read automatically "
                 "if found in same the same dir as extensions.conf; "
                 "set empty to disable")
        return parser

    def handle_args(self, args):
        # No messages to stderr, but do collect the E_APP_MISSING and
        # E_FUNC_MISSING errors.
        aggregator = Aggregator()
        MessageDefManager.muted = True

        # Load func_odbc functions if requested.
        load_func_odbc_functions(args.func_odbc, args.dialplan)

        parser = FileDialplanParser()
        parser.include(args.dialplan)
        dialplan = next(iter(parser))
        del dialplan

        apploader = AppLoader()
        funcloader = FuncLoader()

        self.print_modules_used(apploader, funcloader)
        self.print_load_statements(
            set(apploader.used_modules) |
            set(funcloader.used_modules))
        self.print_unknowns(aggregator)

        return int(
            bool(aggregator.unknown_apps) or
            bool(aggregator.unknown_funcs))

    def print_modules_used(self, apploader, funcloader):
        for what, used_items, used_modules in (
                ('Application', apploader.used_apps, apploader.used_modules),
                ('Function', funcloader.used_funcs, funcloader.used_modules)):
            used_items_per_module = defaultdict(list)
            for item in used_items:
                used_items_per_module[item.module].append(item)

            print('; {} providing modules used:'.format(what))
            for module in used_modules:
                items = used_items_per_module[module][:]
                item_lines = ['']
                while items:
                    next_ = items.pop(0)
                    if item_lines[-1]:
                        item_lines[-1] += ', '
                    formatted = '{}()'.format(next_.name)
                    if len(item_lines[-1]) + len(formatted) > 52:
                        item_lines.append(formatted)
                    else:
                        item_lines[-1] += formatted

                # Output.
                print(';   {:20s}  {}'.format(
                    module, item_lines[0].strip()))
                for item_line in item_lines[1:]:
                    print(';   {:20s}  {}'.format(
                        '', item_line.strip()))
            print(';')

    def print_load_statements(self, all_modules):
        print('; modules.conf')
        for module in sorted(all_modules):
            if module != '<builtin>':
                print('load => {}.so'.format(module))
        print()

    def print_unknowns(self, aggregator):
        if aggregator.unknown_apps:
            print('; WARNING: The following unknown applications were seen:')
            print(';   {}'.format(', '.join(sorted(aggregator.unknown_apps))))
            print(';')
        if aggregator.unknown_funcs:
            print('; WARNING: The following unknown functions were seen:')
            print(';   {}'.format(', '.join(sorted(aggregator.unknown_funcs))))
            print(';')

main = Main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Do sanity checks on func_odbc.conf. Takes 'func_odbc.conf' as argument.
Suppress errors using ALINT_IGNORE env.
"""
from asterisklint import FileFuncOdbcParser
from asterisklint.defines import MessageDefManager
from asterisklint.mainutil import MainBase


class Main(MainBase):
    def create_argparser(self, argparser_class):
        parser = argparser_class(
            description=(
                'Do sanity checks on func_odbc.conf. Suppress comma separated '
                'error classes through the ALINT_IGNORE environment variable. '
                'Returns 1 if any issue was reported.'))
        parser.add_argument(
            'func_odbc', metavar='FUNC_ODBC_CONF', nargs='?',
            default='./func_odbc.conf',
            help="path to func_odbc.conf")
        return parser

    def handle_args(self, args):
        parser = FileFuncOdbcParser()
        parser.include(args.func_odbc)
        for func_odbc_context in parser:
            pass

        # MessageDefManager.raised is a dict of messages ordered by
        # message type. All message types share the same muted flag,
        # so we need only examine the first.
        if any(not i[0].muted for i in MessageDefManager.raised.values()):
            return 1

main = Main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..defines import ErrorDef, WarningDef


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_FUNC_BAD_ARGS(ErrorDef):
        message = "function {func!r} does not take these arguments '{data}'"

    class W_FUNC_BALANCE(WarningDef):
        message = ('function data {data!r} looks like unbalanced '
                   'parentheses/quotes/curlies')



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..defines import ErrorDef, WarningDef


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_APP_ARG_FEW(ErrorDef):
        message = 'too few arguments for app {app!r}, minimum is {min_args}'

    class E_APP_ARG_MANY(ErrorDef):
        message = 'too many arguments for app {app!r}, maximum is {max_args}'

    class E_APP_ARG_BADOPT(ErrorDef):
        message = ('unrecognised options {opts!r} in arg {argno} '
                   'for app {app!r}')

    class E_APP_ARG_DUPEOPT(ErrorDef):
        message = 'duplicate options {opts!r} in arg {argno} for app {app!r}'

    class E_APP_ARG_IFCONST(ErrorDef):
        message = ("apparent constant in If-condition; app {app!r}, "
                   "data '{data}' and cond '{cond}'")

    class E_APP_ARG_IFEMPTY(ErrorDef):
        message = "empty If-condition; app {app!r}, data '{data}'"

    class E_APP_ARG_IFSTYLE(ErrorDef):
        message = ("{app!r} takes the form <cond>?<iftrue>[:<iffalse>] "
                   "but data is '{data}', cond '{cond}', args '''{args}'''")

    class E_APP_ARG_PIPEDELIM(ErrorDef):
        message = ('the application delimiter is now the comma, not '
                   'the pipe; see app {app!r} and data {data!r}')

    class E_APP_ARG_SYNTAX(ErrorDef):
        message = ('generic application syntax error; app {app!r} and '
                   'data {data!r}')

    class E_APP_MISSING(ErrorDef):
        message = 'app {app!r} does not exist, dialplan will halt here!'

    class W_APP_BAD_CASE(WarningDef):
        message = 'app {app!r} does not have the proper Case {proper!r}'

    class W_APP_BALANCE(WarningDef):
        message = ('app data {data!r} looks like unbalanced '
                   'parentheses/quotes/curlies')



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .. import E_FUNC_BAD_ARGS
from ..base import FuncBase


class AUDIOHOOK_INHERIT(FuncBase):
    def __call__(self, data, where):
        if isinstance(data, str) and data not in (
                'MixMonitor', 'Chanspy', 'Volume', 'Speex',
                'pitch_shift', 'JACK_HOOK', 'Mute'):
            E_FUNC_BAD_ARGS(where, func=self.name, data=data)

        super().__call__(data, where)


def register(func_loader):
    func_loader.register(AUDIOHOOK_INHERIT())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import AppBase


class Pickup(AppBase):
    pass


class PickupChan(AppBase):
    pass


class PickupOld1v4(AppBase):
    """
    This is not in any main Asterisk branch. Quickest fix was to add
    here for now.

    NOTE: https://issues.asterisk.org/jira/browse/ASTERISK-26464
    """
    pass


def register(app_loader):
    for app in (
            Pickup,
            PickupChan,
            PickupOld1v4):
        app_loader.register(app())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Show dialplan like Asterisk does with CLI command "dialplan show". Takes
'extensions.conf' as argument.
"""
from asterisklint import FileDialplanParser
from asterisklint.defines import MessageDefManager
from asterisklint.mainutil import MainBase


class Main(MainBase):
    def create_argparser(self, argparser_class):
        parser = argparser_class(
            description=(
                'Shows the dialplan like Asterisk does with the CLI command '
                '"dialplan show". Useful for testing whether asterisklint '
                'parser the input properly.'))
        parser.add_argument(
            'dialplan', metavar='EXTENSIONS_CONF', nargs='?',
            default='./extensions.conf',
            help='path to extensions.conf')
        parser.add_argument(
            '--reverse', action='store_true',
            help="some versions of Asterisk output the dialplan file "
                 "in reverse")
        return parser

    def handle_args(self, args):
        MessageDefManager.muted = True  # no messages to stderr
        parser = FileDialplanParser()
        parser.include(args.dialplan)
        dialplan = next(iter(parser))
        print(dialplan.format_as_dialplan_show(
            reverse=args.reverse))

main = Main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from . import W_FUNC_BALANCE
from ..variable import variable_check_balance


class FuncBase(object):
    @property
    def name(self):
        return self.__class__.__name__

    @property
    def module(self):
        return self.__module__.rsplit('.', 1)[-1]

    def __call__(self, data, where):
        try:
            self.check_balance(data)
        except ValueError:
            W_FUNC_BALANCE(where, data=str(data))

    @staticmethod
    def check_balance(data):
        variable_check_balance(data)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import (
    W_APP_BALANCE,
    AppBase, DelimitedArgsMixin, MinMaxArgsMixin, NoPipeDelimiterMixin,
    VarCondIfStyleApp)


class Gosub(NoPipeDelimiterMixin, MinMaxArgsMixin, DelimitedArgsMixin,
            AppBase):
    @classmethod
    def split_gosub_prio_args(self, prio, where):
        """
        Split off the (<args>) from from <prio>(<args>).
        """
        args = None
        for i, token in enumerate(prio):
            if token == '(':
                if prio[-1] != ')':
                    W_APP_BALANCE(where, data=prio)
                    args = prio[(i + 1):]
                else:
                    args = prio[(i + 1):-1]
                prio = prio[0:i]
                break

        # FYI: The args are parsed by Asterisk using AST_STANDARD_RAW_ARGS.
        return prio, args

    def __init__(self, **kwargs):
        super().__init__(min_args=1, max_args=3, **kwargs)

    def __call__(self, data, where, jump_destinations):
        args = super().__call__(data, where, jump_destinations)

        jumpdest = args[:]
        while len(jumpdest) != 3:
            jumpdest.insert(0, None)
        assert len(jumpdest) == 3
        # Quickly drop the Gosub-ARGn for now.
        jumpdest[2] = Gosub.split_gosub_prio_args(jumpdest[2], where)[0]
        jump_destinations.append(tuple(jumpdest))

        return args


class GosubIf(VarCondIfStyleApp):
    def __call__(self, data, where, jump_destinations):
        cond, iftrue, iffalse = super().__call__(
            data, where, jump_destinations)

        for args in (iftrue, iffalse):
            if args:
                jumpdest = self.separate_args(args)
                while len(jumpdest) != 3:
                    jumpdest.insert(0, None)
                assert len(jumpdest) == 3
                # Quickly drop the Gosub-ARGn for now.
                jumpdest[2] = Gosub.split_gosub_prio_args(
                    jumpdest[2], where)[0]
                jump_destinations.append(tuple(jumpdest))

        return cond, iftrue, iffalse


class Return(AppBase):
    pass


class StackPop(AppBase):
    pass


def register(app_loader):
    for app in (
            Gosub, GosubIf, Return, StackPop):
        app_loader.register(app())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re


class Var(object):
    """
    Var is a nested structure of variables.

    For example::

        "${abc}"    => Var('abc')
        "abc${def}" => Var.join(['abc', Var('abc')])
        "${${abc}}" => Var(Var('abc'))
        "${a${bc}}" => Var(Var.join(['a', Var('bc')]))

    You can call stringify the Var back to their original values::

        str(Var('abc')) => "${abc}"
        ...

    You can also format the Var with variable replacements:

        Var(Var.join(['a', Var('bc')])).format(bc='BC', aBC='foobar') =>
          "foobar"
    """
    @classmethod
    def join(self, variables):
        copy = []
        for var in variables:
            if not isinstance(var, (str, Var)):
                raise TypeError(
                    'Variables can only consist of strings and other '
                    'variables, got {!r}'.format(var))

            if not var:
                # Skip empty vars.
                pass
            elif copy and isinstance(var, str) and isinstance(copy[-1], str):
                # Join consecutive strings together.
                copy[-1] = copy[-1] + var
            else:
                # Append.
                copy.append(var)

        if len(copy) == 0:
            return ''
        if len(copy) == 1:
            return copy[0]

        # Use private setter instead of constructor.
        ret = Var()
        ret._list = copy
        return ret

    def __init__(self, name=None, start=None, length=None):
        self.name = name

    def could_match(self, value):
        """
        Check if this variable could match the value passed if the
        circumstances (the inner variables) are right.

        Examples::

            Var('${number}23').could_match('123') == True
            Var('${user}@${domain}').could_match('somelabel') == False
        """
        if self.name:  # this is a plain variable, it could match anything
            return True

        re_joined = ''.join(
            ['^'] +
            [(re.escape(i) if isinstance(i, str) else '.*') for i in self] +
            ['$'])
        return re.match(re_joined, str(value))

    def format(self, **kwargs):
        if self.name is None:
            ret = []
            for var in self._list:
                if isinstance(var, str):
                    ret.append(var)
                else:
                    ret.append(var.format(**kwargs))
            return ''.join(ret)

        elif isinstance(self.name, Var):
            ret = self.name.format(**kwargs)
            return kwargs[ret]

        else:
            return kwargs[self.name]

    def __iter__(self):
        """
        You may iterate over this thing, as if it where a string. In
        which case you get the literal letters, and variables in
        between.
        """
        return iter(self._get_cached_iter())

    def __getitem__(self, *args, **kwargs):
        """
        Get an item or list of items.

        If a slice is requested (and a list is returned), we attempt to
        join the list using strjoin. Examples::

            ['A', 'B', 'C', <Var>, 'D'] ==> ['ABC', <Var>, 'D']
            ['A', 'B', 'C'] ==> 'ABC'
            ['A'] ==> 'A'
            [<Var>] ==> <Var>
            [] ==> []

        This should never be a problem because you'll be iterating over
        the results anyway.
        """
        item = self._get_cached_iter().__getitem__(*args, **kwargs)
        if isinstance(item, list):
            ret = list(strjoin(item))  # joins consecutive strings together
            if len(ret) == 1:
                return ret[0]
            return ret
        return item

    def __len__(self):
        return len(self._get_cached_iter())

    def _get_cached_iter(self):
        if not hasattr(self, '_cached_iter'):
            ret = []
            if self.name is None:
                for var in self._list:
                    for inner in var:
                        ret.append(inner)
            else:
                ret.append(self)
            self._cached_iter = ret
        return self._cached_iter

    def __eq__(self, other):
        if not isinstance(other, Var):
            return False

        if (self.name, other.name) == (None, None):
            if len(self._list) != len(other._list):
                return False
            return all(self._list[i] == other._list[i]
                       for i in range(len(self._list)))

        if None in (self.name, other.name):
            return False

        return self.name == other.name

    def __str__(self):
        if self.name:
            return '${{{}}}'.format(self.name)
        return ''.join(str(i) for i in self._list)

    def split(self, token=':', maxsplit=-1):
        # In case we're an atom, pretend we're a list.
        if self.name:
            list_ = [self.name]  # [name]  # (start, length)
        else:
            list_ = self._list   # [contents...]  # (start, length)

        assert (
            (token == ':' and maxsplit == -1) or    # ${var:start:len}
            (token == '=' and maxsplit == 1) or     # Set(X=Y)
            (token == '(' and maxsplit == 1)        # FUNC(x)
        ), (token, maxsplit)  # unexpected split token..

        ret = self._split(list_, token, maxsplit)
        return ret

    def _split(self, source, token, maxsplit):
        # Split it up by token, so we group the values into subgroups.
        split_up = [[]]
        maxsplit_done = 0

        for item in source:
            if not isinstance(item, str):
                split_up[-1].append(item)
            elif token not in item or (maxsplit == maxsplit_done):
                split_up[-1].append(item)
            else:
                if maxsplit == -1:
                    subitems = item.split(token, maxsplit)
                else:
                    subitems = item.split(token, maxsplit - maxsplit_done)

                while len(subitems) > 1:
                    first = subitems.pop(0)
                    if first:
                        split_up[-1].append(first)
                    split_up.append([])
                    maxsplit_done += 1

                if subitems[0]:
                    split_up[-1].append(subitems[0])

        # Join the subgroups back into variables.
        for i, list_ in enumerate(split_up):
            if len(list_) == 0:
                split_up[i] = ''
            elif len(list_) > 1:
                split_up[i] = Var.join(split_up[i])
            else:
                split_up[i] = split_up[i][0]

        return split_up

    def strip(self):
        if self.name:
            return self  # no (deep)copy?

        new_list = self._list[:]
        if isinstance(new_list[0], str) and new_list[0].strip() == '':
            new_list = new_list[1:]
        if (new_list and isinstance(new_list[-1], str) and
                new_list[-1].strip() == ''):
            new_list = new_list[0:-1]

        if len(new_list) == len(self._list):
            return self  # no (deep)copy?
        if len(new_list) == 1:
            return new_list[0]

        v = Var()
        v._list = new_list  # no deepcopy..
        return v


class SliceMixin(object):
    def __init__(self, *args, start=None, length=None, **kwargs):
        assert start is not None
        super().__init__(*args, **kwargs)
        self.start = start
        self.length = length
        if length is not None:
            assert length != 0
            if length < 0:
                self._endpos = length
            elif self.start < 0:
                self._endpos = self.start + length
                assert self._endpos < 0
            else:
                self._endpos = self.start + length

    def format(self, **kwargs):
        value = super().format(**kwargs)
        if self.length:
            return value[self.start:self._endpos]
        return value[self.start:]

    def __str__(self):
        if self.length:
            return '${{{}:{}:{}}}'.format(
                self.name, self.start, self.length)
        return '${{{}:{}}}'.format(
            self.name, self.start)


class VarSlice(SliceMixin, Var):
    def __init__(self, name=None, start=None, length=None):
        assert name is not None
        super().__init__(name=name, start=start, length=length)


class VarDynSlice(Var):
    def __init__(self, name=None, start=None, length=None):
        assert name is not None
        super().__init__(name=name)
        self.start = start
        self.length = length

    def __str__(self):
        if self.length:
            return '${{{}:{}:{}}}'.format(
                self.name, self.start, self.length)
        return '${{{}:{}}}'.format(
            self.name, self.start)


def strjoin(list_of_items_and_strings):
    """
    Joins all consecutive items that are strings together.

    E.g.: [1, 2, 'a', 'b', 'c', 3, 'd', 'e'] ==> [1, 2, 'abc', 3, 'de']
    """
    cache = []
    for ch in list_of_items_and_strings:
        if isinstance(ch, str):
            cache.append(ch)
        else:
            if cache:
                yield ''.join(cache)
                cache = []
            yield ch
    if cache:
        yield ''.join(cache)


def variable_check_balance(data):
    """
    This function is a simple check that looks sane, but is not
    implemented like this in Asterisk. It may warn you of badly
    placed delimiters, but you shouldn't use this to interpret
    variables.

    For that, see separate_args which is modeled after the function
    as implemented in Asterisk.
    """
    arr = ['X']

    for char in data:
        if isinstance(char, Var):
            # It's possible that we're looping over a Var variable.
            # In that case we'll have to assume the Var itself
            # contains no separators that we might be interested in.
            # E.g. we expect you do *not* do this:
            # ``Set(var=mailbox@context,s)``
            # ``VoiceMail(${var})``
            # But we expect you to do this:
            # ``Set(mailbox=mailbox@context)``
            # ``VoiceMail(${mailbox},s)``
            # Same goes for functions.
            pass  # skip char
        elif char == '"':
            if arr[-1] == '"':
                arr.pop()
            elif arr[-1] == "'":
                pass
            else:
                arr.append('"')
        elif char == "'":
            if arr[-1] == "'":
                arr.pop()
            elif arr[-1] == '"':
                pass
            else:
                arr.append("'")
        elif char in '({[':
            if arr[-1] in '\'"':
                pass
            else:
                arr.append(char)
        elif char in ')}]':
            left = '({['[')}]'.index(char)]
            if arr[-1] in '\'"':
                pass
            else:
                if arr[-1] == left:
                    arr.pop()
                else:
                    raise ValueError(''.join(arr[1:]))

    if arr != ['X']:
        raise ValueError(''.join(arr[1:]))



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Do sanity checks on dialplan. Takes 'extensions.conf' as argument.
Suppress errors using ALINT_IGNORE env.
"""
from asterisklint import FileDialplanParser
from asterisklint.defines import MessageDefManager
from asterisklint.mainutil import (
    MainBase, UniqueStore, load_func_odbc_functions)


class Main(MainBase):
    def create_argparser(self, argparser_class):
        parser = argparser_class(
            description=(
                'Do sanity checks on dialplan. Suppress comma separated '
                'error classes through the ALINT_IGNORE environment variable. '
                'Returns 1 if any issue was reported.'))
        parser.add_argument(
            'dialplan', metavar='EXTENSIONS_CONF', nargs='?',
            default='./extensions.conf',
            help='path to extensions.conf')
        parser.add_argument(
            '--func-odbc', metavar='FUNC_ODBC_CONF', action=UniqueStore,
            help="path to func_odbc.conf, will be read automatically if found "
                 "in same the same dir as extensions.conf; "
                 "set empty to disable")
        return parser

    def handle_args(self, args):
        # Load func_odbc functions if requested.
        load_func_odbc_functions(args.func_odbc, args.dialplan)

        parser = FileDialplanParser()
        parser.include(args.dialplan)
        dialplan = next(iter(parser))
        dialplan.walk_jump_destinations()
        del dialplan

        # MessageDefManager.raised is a dict of messages ordered by message
        # type. All message types share the same muted flag, so we need only
        # examine the first.
        if any(not i[0].muted for i in MessageDefManager.raised.values()):
            return 1

main = Main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Report similarly named contexts, labels and variables. Takes
'extensions.conf' as argument. All parse errors are suppressed.
"""
from collections import defaultdict

from asterisklint import FileDialplanParser
from asterisklint.defines import MessageDefManager
from asterisklint.varfun import VarLoader
from asterisklint.mainutil import MainBase

try:
    import editdistance  # https://pypi.python.org/pypi/editdistance
except ImportError:
    editdistance = None


class Main(MainBase):
    def create_argparser(self, argparser_class):
        parser = argparser_class(
            description=(
                'Report similarly named contexts, labels and variables. '
                'All parse errors are suppressed. Returns 1 if any potential '
                'issue was reported.'))
        parser.add_argument(
            '-v', '--verbose', action='store_true',
            help='list all identifiers first before reporting similarities')
        parser.add_argument(
            'dialplan', metavar='EXTENSIONS_CONF', nargs='?',
            default='./extensions.conf',
            help='path to extensions.conf')
        return parser

    def handle_args(self, args):
        MessageDefManager.muted = True

        loader = VarLoader()
        parser = FileDialplanParser()
        parser.include(args.dialplan)
        dialplan = next(iter(parser))

        contexts_by_name = list(sorted(
            (context.name for context in dialplan.contexts),
            key=(lambda x: x.lower())))
        # TODO: dialplan.all_labels is not a public interface..
        labels_by_name = list(sorted(
            dialplan.all_labels, key=(lambda x: x.lower())))
        # TODO: loader._variables is *not* a public interface..
        varlist_by_name = list(sorted(
            loader._variables.items(), key=(lambda x: x[0].lower())))

        if args.verbose:
            self.print_contexts(contexts_by_name)
            self.print_labels(labels_by_name)
            self.print_variables(varlist_by_name)

        # Calculate Levenshtein distance if available. Complain if
        # module wasn't loaded and the user did not ask for verbose
        # printing either.
        if editdistance:
            identifiers = (
                set(contexts_by_name) |
                set(labels_by_name) |
                set([i[0] for i in varlist_by_name]))
            ret = self.print_distance(identifiers)
        elif args.verbose:
            ret = 0
        else:
            raise ImportError(
                'Loading editdistance failed. Using this command without '
                'the editdistance and without verbose mode is a no-op.')

        return ret

    def print_contexts(self, contexts_by_name):
        print('Contexts encountered:')
        for context in contexts_by_name:
            print('  {}'.format(context))
        print()

    def print_labels(self, labels_by_name):
        print('Labels encountered:')
        for label in labels_by_name:
            print('  {}'.format(label))
        print()

    def print_variables(self, varlist_by_name):
        print('Variables encountered:')
        for variable, occurrences in varlist_by_name:
            print('  {:32}  [{} times in {} files]'.format(
                variable, len(occurrences),
                len(set(i.filename for i in occurrences))))
        print()

    def print_distance(self, identifiers):
        id_by_name = sorted(
            identifiers, key=(lambda x: x.lower()))
        id_by_len = sorted(
            identifiers, key=(lambda x: (len(x), x.lower())))

        similar = defaultdict(set)
        for id_list in (id_by_name, id_by_len):
            prev = None
            for cur in id_list:
                if prev:
                    lodiff = editdistance.eval(prev.lower(), cur.lower())
                    if (lodiff == 0 and (
                            (prev.isupper() and cur.islower()) or
                            (prev.islower() and cur.isupper()))):
                        pass
                    elif (prev.startswith('ARG') and
                            cur.startswith('ARG') and
                            prev[3:].isdigit() and
                            cur[3:].isdigit()):
                        # ARG1..n as passed to a Gosub() routine.
                        pass
                    elif (lodiff <= 2 and (lodiff + 1) < len(prev) and
                            (lodiff + 1) < len(cur)):
                        similar[prev].add(cur)
                        similar[cur].add(prev)
                prev = cur

        if similar:
            print('Identifiers with similar names include:')
            for identifier in sorted(
                    similar.keys(), key=(lambda x: x.lower())):
                similar_to = ', '.join(sorted(similar[identifier]))
                print('  {:32}  [similar to: {}]'.format(
                    identifier, similar_to))
            print()
            return 1

main = Main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2017  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Expand func_odbc.conf into wrapped/readable/annotated form. Takes
'func_odbc.conf' as argument.
"""
from asterisklint import FileFuncOdbcParser
from asterisklint.defines import MessageDefManager
from asterisklint.mainutil import MainBase


class Main(MainBase):
    def create_argparser(self, argparser_class):
        parser = argparser_class(
            description=(
                'Write the func_odbc SQL statements in a '
                'wrapped/readable/annotated form. In the future we may '
                'add a func_odbc-collapse to collapse the annotated form '
                'again.'))  # so we want output resembling func_odbc.conf
        parser.add_argument(
            'func_odbc', metavar='FUNC_ODBC_CONF', nargs='?',
            default='./func_odbc.conf',
            help="path to func_odbc.conf")
        return parser

    def handle_args(self, args):
        # No messages to stderr.
        MessageDefManager.muted = True

        parser = FileFuncOdbcParser()
        parser.include(args.func_odbc)
        print('-- BEWARE: This output format is not standardized yet')
        print('-- NOT FOR MACHINE CONSUMPTION')
        print('')

        for func_odbc_context in parser:
            print('-- {}'.format(func_odbc_context.get_function_name()))
            print('-- .')
            for key, query in sorted(func_odbc_context.get_queries().items()):
                print('-- FUNCTION TYPE: {}'.format(key))
                if key == 'read':
                    columns = query.select_columns()
                    if columns:
                        print('-- COLUMNS:')
                        for idx, column in enumerate(columns):
                            print('-- {:2d}. {}'.format(idx + 1, column))
                formatted = query.format_sql()
                if not formatted.endswith(';'):
                    formatted += ';'
                print(formatted)
            print()

main = Main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import FuncBase


class QUEUE_EXISTS(FuncBase):
    pass


class QUEUE_MEMBER(FuncBase):
    pass


class QUEUE_MEMBER_COUNT(FuncBase):
    pass


class QUEUE_MEMBER_LIST(FuncBase):
    pass


class QUEUE_MEMBER_PENALTY(FuncBase):
    pass


class QUEUE_VARIABLES(FuncBase):
    pass


class QUEUE_WAITING_COUNT(FuncBase):
    pass


def register(func_loader):
    for func in (
            QUEUE_EXISTS, QUEUE_MEMBER, QUEUE_MEMBER_COUNT,
            QUEUE_MEMBER_LIST, QUEUE_MEMBER_PENALTY,
            QUEUE_VARIABLES, QUEUE_WAITING_COUNT):
        func_loader.register(func())
